"""
.. module:: variables
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`AKIT_VARIABLES` object which is used store the environment variables.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import logging
import os
import sys
import uuid

from datetime import datetime
from enum import Enum

from akit.conversions import string_to_bool
from akit.xtime import parse_datetime

environ = os.environ

THIS_DIR = os.path.dirname(__file__)

LOG_LEVEL_NAMES = [
    "NOTSET",
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "QUIET"
]

LOG_LEVEL_VALUES = {
    "NOTSET": logging.NOTSET,
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
    "QUIET": 100
}

def normalize_variable_whitespace(lval):
    lval = lval.strip().replace("/t", " ")
    while lval.find("  ") > -1:
        lval = lval.replace("  ", " ")
    return lval

class JOB_TYPES:
    UNKNOWN = "unknown"
    COMMAND = "command"
    CONSOLE = "console"
    TESTRUN = "testrun"
    ORCHESTRATION = "orchestration"
    SERVICE = "service"


class ActivationProfile(str, Enum):
    Command = "command"
    Console = "console"
    Orchestration = "orchestration"
    Service = "service"
    TestRun = "testrun"


class AKIT_VARIABLES:
    """
        Container for all the configuration variables that can be passed via environmental variables.
    """

    AKIT_DIR = os.path.abspath(os.path.join(THIS_DIR, ".."))

    AKIT_ACTIVATION_PROFILE = None

    AKIT_INTERACTIVE_CONSOLE = False
    if "AKIT_INTERACTIVE_CONSOLE" in environ:
        AKIT_INTERACTIVE_CONSOLE = string_to_bool(environ["AKIT_INTERACTIVE_CONSOLE"].lower())

    AKIT_APOD_NAME = "unknown"
    if "AKIT_APOD_NAME" in environ:
        AKIT_APOD_NAME = environ["AKIT_APOD_NAME"]

    AKIT_BUILD_BRANCH = "unknown"
    if "AKIT_BUILD_BRANCH" in environ:
        AKIT_BUILD_BRANCH = environ["AKIT_BUILD_BRANCH"]

    AKIT_BUILD_NAME = "unknown"
    if "AKIT_BUILD_NAME" in environ:
        AKIT_BUILD_NAME = environ["AKIT_BUILD_NAME"]

    AKIT_BUILD_URL = "unknown"
    if "AKIT_BUILD_URL" in environ:
        AKIT_BUILD_URL = environ["AKIT_BUILD_URL"]

    AKIT_JOB_INITIATOR = "unknown"
    if "AKIT_JOB_INITIATOR" in environ:
        AKIT_JOB_INITIATOR = environ["AKIT_JOB_OWNER"]

    AKIT_JOB_LABEL = "unknown"
    if "AKIT_JOB_LABEL" in environ:
        AKIT_JOB_LABEL = environ["AKIT_JOB_LABEL"]

    AKIT_JOB_NAME = "unknown"
    if "AKIT_JOB_NAME" in environ:
        AKIT_JOB_NAME = environ["AKIT_JOB_NAME"]

    AKIT_JOB_OWNER = "unknown"
    if "AKIT_JOB_OWNER" in environ:
        AKIT_JOB_OWNER = environ["AKIT_JOB_OWNER"]

    AKIT_LOG_LEVEL_CONSOLE = None
    if "AKIT_LOG_LEVEL_CONSOLE" in environ:
        AKIT_LOG_LEVEL_CONSOLE = environ["AKIT_LOG_LEVEL_CONSOLE"].upper()

    AKIT_BREAKPOINTS = None
    if "AKIT_BREAKPOINTS" in environ:
        breakpoints = []
        for bkpnt in environ["AKIT_BREAKPOINTS"].lower().split(","):
            breakpoints.append(bkpnt)
        AKIT_BREAKPOINTS = breakpoints

    AKIT_DEBUGGER = None
    if "AKIT_DEBUGGER" in environ:
        AKIT_DEBUGGER = environ["AKIT_DEBUGGER"].lower()

    AKIT_LOG_LEVEL_FILE = None
    if "AKIT_LOG_LEVEL_FILE" in environ:
        AKIT_LOG_LEVEL_FILE = environ["AKIT_LOG_LEVEL_FILE"].upper()

    AKIT_BUILD_FLAVOR = "unknown"
    if "AKIT_BUILD_FLAVOR" in environ:
        AKIT_BUILD_FLAVOR = environ["AKIT_BUILD_FLAVOR"]

    AKIT_JOBTYPE = "unknown"
    if "AKIT_JOBTYPE" in environ:
        AKIT_JOBTYPE = environ["AKIT_JOBTYPE"]

    AKIT_CONFIG_EXTENSION_POINTS_MODULE = "akit.extensionpoints"
    if "AKIT_CONFIG_EXTENSION_POINTS_MODULE" in environ:
        AKIT_CONFIG_EXTENSION_POINTS_MODULE = environ["AKIT_CONFIG_EXTENSION_POINTS_MODULE"]

    AKIT_RUNID = None
    if "AKIT_RUNID" in environ:
        AKIT_RUNID = environ["AKIT_RUNID"]
    else:
        AKIT_RUNID = str(uuid.uuid4())

    # AKIT_SERVICE_NAME is always set by environment variable
    # as a service starts up.
    AKIT_SERVICE_NAME = None

    AKIT_STARTTIME = None
    if "AKIT_STARTTIME" in environ:
        starttime = parse_datetime(environ["AKIT_STARTTIME"])
        AKIT_STARTTIME = starttime
    elif AKIT_STARTTIME is None:
        AKIT_STARTTIME = datetime.now()

    AKIT_HOME_DIRECTORY = os.path.expanduser("~/akit")
    if "AKIT_HOME_DIRECTORY" in environ:
        AKIT_HOME_DIRECTORY = environ["AKIT_HOME_DIRECTORY"]

    AKIT_CONFIG_DIRECTORY = os.path.join(AKIT_HOME_DIRECTORY, "config")
    if "AKIT_CONFIG_DIRECTORY" in environ:
        AKIT_CONFIG_DIRECTORY = environ["AKIT_CONFIG_DIRECTORY"]

    AKIT_CONFIG_CREDENTIALS = os.path.join(AKIT_CONFIG_DIRECTORY, "credentials.yaml")
    if "AKIT_CONFIG_CREDENTIALS" in environ:
        AKIT_CONFIG_CREDENTIALS = environ["AKIT_CONFIG_CREDENTIALS"].upper()

    AKIT_CONFIG_LANDSCAPE_NAME = "default-landscape"
    if "AKIT_CONFIG_LANDSCAPE_NAME" in environ:
        AKIT_CONFIG_LANDSCAPE_NAME = environ["AKIT_CONFIG_LANDSCAPE_NAME"]

    AKIT_CONFIG_LANDSCAPE = os.path.join(AKIT_CONFIG_DIRECTORY, "landscapes", AKIT_CONFIG_LANDSCAPE_NAME + ".yaml")
    if "AKIT_CONFIG_LANDSCAPE" in environ:
        AKIT_CONFIG_LANDSCAPE = environ["AKIT_CONFIG_LANDSCAPE"]

    AKIT_CONFIG_RUNTIME_NAME = "default-runtime"
    if "AKIT_CONFIG_RUNTIME_NAME" in environ:
        AKIT_CONFIG_RUNTIME_NAME = environ["AKIT_CONFIG_RUNTIME_NAME"]

    AKIT_CONFIG_RUNTIME_SEARCH_PATH = None
    if "AKIT_CONFIG_RUNTIME_SEARCH_PATH" in environ:
        AKIT_CONFIG_RUNTIME_SEARCH_PATH = environ["AKIT_CONFIG_RUNTIME_SEARCH_PATH"]
    else:
        AKIT_CONFIG_RUNTIME_SEARCH_PATH = os.path.join(AKIT_CONFIG_DIRECTORY, "runtimes")

    AKIT_CONFIG_RUNTIME = None
    if "AKIT_CONFIG_RUNTIME" in environ:
        AKIT_CONFIG_RUNTIME = environ["AKIT_CONFIG_RUNTIME"]
    else:
        runtime_search_path_list = AKIT_CONFIG_RUNTIME_SEARCH_PATH.split(":")
        for spath in runtime_search_path_list:
            check_path = os.path.join(spath, AKIT_CONFIG_RUNTIME_NAME + ".yaml")
            if os.path.exists(check_path):
                AKIT_CONFIG_RUNTIME = check_path
                break
    
    # If we still did not find a runtime file, set it to the default location
    # the configuration loading code can warn if the file does not exist.
    if AKIT_CONFIG_RUNTIME is None:
        AKIT_CONFIG_RUNTIME = os.path.join(AKIT_CONFIG_DIRECTORY, "runtimes", AKIT_CONFIG_RUNTIME_NAME + ".yaml")

    AKIT_CONFIG_TOPOLOGY_NAME = "default-topology"
    if "AKIT_CONFIG_TOPOLOGY_NAME" in environ:
        AKIT_CONFIG_TOPOLOGY_NAME = environ["AKIT_CONFIG_TOPOLOGY_NAME"]

    AKIT_CONFIG_TOPOLOGY = os.path.join(AKIT_CONFIG_DIRECTORY, "topologies", AKIT_CONFIG_TOPOLOGY_NAME + ".yaml")
    if "AKIT_CONFIG_TOPOLOGY" in environ:
        AKIT_CONFIG_TOPOLOGY = environ["AKIT_CONFIG_TOPOLOGY"]

    AKIT_CONFIG_USER = os.path.join(AKIT_CONFIG_DIRECTORY, "user.yaml")
    if "AKIT_CONFIG_USER" in environ:
        AKIT_CONFIG_USER = environ["AKIT_CONFIG_USER"]

    AKIT_OUTPUT_DIRECTORY = None
    if "AKIT_OUTPUT_DIRECTORY" in environ:
        AKIT_OUTPUT_DIRECTORY = environ["AKIT_OUTPUT_DIRECTORY"]

    AKIT_SKIP_DEVICES = None
    if "AKIT_SKIP_DEVICES" in environ:
        AKIT_SKIP_DEVICES = environ["AKIT_SKIP_DEVICES"]

    AKIT_RESULTS_STATIC_RESOURCE_DEST_DIR = os.path.join(AKIT_HOME_DIRECTORY, "results", "static")
    if "AKIT_RESULTS_STATIC_RESOURCE_DEST_DIR" in environ:
        AKIT_RESULTS_STATIC_RESOURCE_DEST_DIR = environ["AKIT_RESULTS_STATIC_RESOURCE_DEST_DIR"]
    
    AKIT_RESULTS_STATIC_RESOURCE_SRC_DIR = os.path.join(AKIT_DIR, "templates", "static")
    if "AKIT_RESULTS_STATIC_RESOURCE_SRC_DIR" in environ:
        AKIT_RESULTS_STATIC_RESOURCE_SRC_DIR = environ["AKIT_RESULTS_STATIC_RESOURCE_SRC_DIR"]

    AKIT_RESULTS_HTML_TEMPLATE = None
    if "AKIT_RESULTS_HTML_TEMPLATE" in environ:
        AKIT_RESULTS_HTML_TEMPLATE = environ["AKIT_RESULTS_HTML_TEMPLATE"]

    AKIT_TESTROOT = None
    if "AKIT_TESTROOT" in environ:
        AKIT_TESTROOT = environ["AKIT_TESTROOT"]

    AKIT_TRACEBACK_POLICY_OVERRIDE = None
    if "AKIT_TRACEBACK_POLICY_OVERRIDE" in environ:
        AKIT_TRACEBACK_POLICY_OVERRIDE = environ["AKIT_TRACEBACK_POLICY_OVERRIDE"]

    AKIT_UPNP_SCAN_INTEGRATION_BASE = None
    if "AKIT_UPNP_SCAN_INTEGRATION_BASE" in environ:
        AKIT_UPNP_SCAN_INTEGRATION_BASE = environ["AKIT_UPNP_SCAN_INTEGRATION_BASE"]

    AKIT_UPNP_EXTENSIONS_INTEGRATION_BASE = None
    if "AKIT_UPNP_EXTENSIONS_INTEGRATION_BASE" in environ:
        AKIT_UPNP_EXTENSIONS_INTEGRATION_BASE = environ["AKIT_UPNP_EXTENSIONS_INTEGRATION_BASE"]

    AKIT_UPNP_DYN_EXTENSIONS_MODULE = None
    if "AKIT_UPNP_DYN_EXTENSIONS_MODULE" in environ:
        AKIT_UPNP_DYN_EXTENSIONS_MODULE = environ["AKIT_UPNP_DYN_EXTENSIONS_MODULE"]
    
    AKIT_TIMETRAVEL = False
    if "AKIT_TIMETRAVEL" in environ:
        AKIT_TIMETRAVEL = environ["AKIT_TIMETRAVEL"].lower()

    AKIT_TIMEPORTALS = None
    if "AKIT_TIMEPORTALS" in environ:
        AKIT_TIMEPORTALS = string_to_bool(environ["AKIT_TIMEPORTALS"].lower())

def extend_path(dir_to_add):
    """
        Extends the PYTHONPATH in the current python process and also modifies
        'PYTHONPATH' so the child processes will also see inherit the extension
        of 'PYTHONPATH'.
    """
    found = False

    for nxt_item in sys.path:
        nxt_item = nxt_item.rstrip(os.sep)
        dir_to_add = dir_to_add.rstrip(os.sep)
        if nxt_item == dir_to_add:
            found = True
            break

    if not found:
        sys.path.insert(0, dir_to_add)
        if "PYTHONPATH" in os.environ:
            os.environ["PYTHONPATH"] = dir_to_add + os.pathsep + os.environ["PYTHONPATH"]
        else:
            os.environ["PYTHONPATH"] = dir_to_add

    return



"""
.. module:: sshcoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains a SshPoolCoordinatorIntegration object to use for working with the computer nodes via SSH

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, Generator, List, Tuple, TYPE_CHECKING

from akit.exceptions import AKitConfigurationError, AKitSemanticError
from akit.coupling.coordinatorcoupling import CoordinatorCoupling
from akit.interop.coordinators.sshpoolcoordinator import SshPoolCoordinator

from akit.testing import testplus

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class SshPoolCoordinatorIntegration(CoordinatorCoupling):
    """
        The SshPoolCoordinatorIntegration handle the requirement registration for the SSH coordinator.
    """

    pathbase = "/ssh"

    def __init__(self, *args, **kwargs):
        """
            The default contructor for an :class:`SshPoolCoordinatorIntegration`.
        """
        super(SshPoolCoordinatorIntegration, self).__init__(*args, **kwargs)
        if self.pathbase is None:
            raise ValueError("The 'pathbase' class member variable must be set to a unique name for each integration class type.")

        self.context.insert(self.pathbase, self)
        return

    @classmethod
    def attach_to_environment(cls, constraints: Dict={}):
        """
            This API is called so that the IntegrationCoupling can process configuration information.  The :class:`IntegrationCoupling`
            will verify that it has a valid environment and configuration to run in.

            :raises :class:`akit.exceptions.AKitMissingConfigError`, :class:`akit.exceptions.AKitInvalidConfigError`:
        """
        resources_acquired = False

        ssh_device_list = cls.landscape.get_ssh_device_list()

        if len(ssh_device_list) > 0:
            resources_acquired = True

        if resources_acquired:
            cls.landscape.activate_integration_point("coordinator/ssh", cls.create_coordinator)

        return

    @classmethod
    def attach_to_framework(cls, landscape: "Landscape"):
        """
            This API is called so that the IntegrationCoupling can attach to the test framework and participate with
            registration processes.  This allows the framework to ignore the bring-up of couplings that are not being
            included by a test.
        """
        super(SshPoolCoordinatorIntegration, cls).attach_to_framework(landscape)
        cls.landscape.register_integration_point("coordinator/ssh", cls)
        return

    @classmethod
    def collect_resources(cls):
        """
            This API is called so the `IntegrationCoupling` can connect with a resource management
            system and gain access to the resources required for the automation run.

            :raises :class:`akit.exceptions.AKitResourceError`:
        """
        return

    @classmethod
    def create_coordinator(cls, landscape: "Landscape") -> object:
        """
            This API is called so that the landscape can create a coordinator for a given integration role.
        """
        cls.coordinator = SshPoolCoordinator(landscape)
        return cls.coordinator

    @classmethod
    def declare_precedence(cls) -> int:
        """
            This API is called so that the IntegrationCoupling can declare an ordinal precedence that should be
            utilized for bringing up its integration state.
        """
        # We need to call the base class, it sets the 'logger' member
        super(SshPoolCoordinatorIntegration, cls).declare_precedence()
        return

    @classmethod
    def diagnostic(cls, label: str, level: int, diag_folder: str):
        """
            The API is called by the :class:`akit.sequencer.Sequencer` object when the automation sequencer is
            building out a diagnostic package at a diagnostic point in the automation sequence.  Example diagnostic
            points are:

            * pre-run
            * post-run

            Each diagnostic package has its own storage location so derived :class:`akit.scope.ScopeCoupling` objects
            can simply write to their specified output folder.

            :param label: The label associated with this diagnostic.
            :param level: The maximum diagnostic level to run dianostics for.
            :param diag_folder: The output folder path where the diagnostic information should be written.
        """
        return

    @classmethod
    def establish_connectivity(cls, allow_missing_devices: bool=False) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish connectivity with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """

        ssh_device_list = cls.landscape.get_ssh_device_list()

        if len(ssh_device_list) == 0:
            raise AKitSemanticError("We should have not been called if no SSH devices are available.")

        upnp_coord = cls.landscape._internal_get_upnp_coord()

        ssh_config_errors, matching_device_results, missing_device_results = cls.coordinator.attach_to_devices(
            ssh_device_list, upnp_coord=upnp_coord)

        ssh_scan_results = {
            "ssh": {
                "matching_devices": matching_device_results,
                "missing_devices": missing_device_results
            }
        }

        return (ssh_config_errors, ssh_scan_results)

    @classmethod
    def establish_presence(cls) -> Tuple[List[str], dict]:
        """
            This API is called so the `IntegrationCoupling` can establish presence with any compute or storage
            resources.

            :returns: A tuple with a list of error messages for failed connections and dict of connectivity
                      reports for devices devices based on the coordinator.
        """
        return

@testplus.integration()
def ssh_coordinator_integration() -> Generator[SshPoolCoordinatorIntegration, None, None]:
    yield SshPoolCoordinatorIntegration()



"""
.. module:: timeouts
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains framework timeout constants.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from typing import List, Optional

from datetime import datetime, timedelta

from akit.exceptions import AKitTimeoutError

DEFAULT_WAIT_DELAY = 0
DEFAULT_WAIT_INTERVAL = 5
DEFAULT_WAIT_TIMEOUT = 60

DEFAULT_SSH_COMMAND_TIMEOUT = 30

DEFAULT_WHATFOR_TEMPLATE = "Timeout waiting for {}"

MSG_TEMPL_TIME_COMPONENTS = "    timeout={} start_time={}, end_time={} now_time={} time_diff={}"

class TimeoutState:
    TimedOut = -1
    NotStarted = 0
    Tracking = 1
    Completed = 2
    Running = 3

class TimeoutContext:
    """
        The :class:`TimeoutContext` object is used to store the context used to track the timeout
        of a single operations or multiple consecutive operations.
    """
    def __init__(self, timeout: float, interval: float=0, delay: float=0, doevery: int=-1, what_for: Optional[str]=None):
        self._timeout = timeout
        self._interval = interval
        self._delay = delay
        self._doevery = doevery
        self._what_for = what_for

        self._loop_counter = 0
        self._now_time = None
        self._start_time = None
        self._end_time = None
        self._final_attempt = False
        self._wait_state = TimeoutState.NotStarted
        return

    @property
    def completed(self) -> bool:
        """
            Indicates the wait was successfully completed.
        """
        return self._wait_state == TimeoutState.Completed

    @property
    def end_time(self) -> datetime:
        """
            Property to retreive the current endtime
        """
        return self._end_time

    @property
    def delay(self) -> float:
        """
            Property for retreiving the delay value.
        """
        return self._delay

    @property
    def final_attempt(self) -> bool:
        """
            Property for retreiving the final_attempt marker and for monitoring and debugging
            calls to look at the final attempt marker.
        """
        return self._final_attempt

    @property
    def has_timed_out(self) -> bool:
        """
            Property indicating if the wait context reached its timeout while running.
        """
        htoval = self._wait_state != TimeoutState.Completed and self._now_time > self._end_time
        return htoval

    @property
    def interval(self) -> float:
        """
            Property for retreiving the interval value.
        """
        return self._interval

    @property
    def is_do_every_interval(self):
        dei = False
        if self._doevery > -1:
            dei = self._loop_counter % self._doevery == 0
        return dei

    @property
    def timeout(self) -> float:
        """
            Property for retreiving the timeout value.
        """
        return self._timeout

    @property
    def wait_state(self) -> TimeoutState:
        """
            Property to return the current wait state of the wait context.
        """
        return self._wait_state

    @property
    def what_for(self) -> str:
        """
            Property for retreiving the what_for value.
        """
        return self._what_for

    def continue_waiting(self):
        """
            Reset the wait context to a waiting state so we can continue waiting.
        """
        self._wait_state = TimeoutState.Running
        return

    def create_timeout(self, what_for: Optional[str]=None, detail: Optional[List[str]]=None, mark_timeout: Optional[bool]=True) -> AKitTimeoutError:
        """
            Helper method used to create detail :class:`AKitTimeoutError` exceptions
            that can be raised in the context of the looper method. 
        """
        if what_for is None:
            what_for = self._what_for

        err_msg = self.format_timeout_message(what_for, detail=detail)
        err_inst = AKitTimeoutError(err_msg)

        if mark_timeout:
            self.mark_timeout()

        return err_inst

    def extend_timeout(self, seconds: float):
        """
            Extend the timeout of the current wait context by the specified number of seconds.

            :param seconds: The time in seconds to extend the wait period.
        """
        self._end_time = self._end_time + timedelta(seconds=seconds)
        self._wait_state = TimeoutState.Running
        self._final_attempt = False
        return

    def format_timeout_message(self, what_for: str, detail: Optional[List[str]]=None) -> str:
        """
            Helper method used to create format a detailed error message for reporting a timeout condition.
        """
        diff_time = self._now_time - self._start_time
        err_msg_lines = [
            "Timeout waiting for {}:".format(what_for),
            MSG_TEMPL_TIME_COMPONENTS.format(self._timeout, self._start_time, self._end_time, self._now_time, diff_time),
        ]

        if detail is not None:
            err_msg_lines.extend(detail)

        err_msg = os.linesep.join(err_msg_lines)
        return err_msg

    def mark_begin(self):
        """
            Mark the wait context as running.
        """
        self._now_time = datetime.now()
        self._start_time = self._now_time
        self._end_time = self._start_time + timedelta(seconds=self._timeout)
        self._wait_state = TimeoutState.Running
        return

    def mark_complete(self):
        """
            Mark the wait context as complete.
        """
        self._wait_state = TimeoutState.Completed
        return

    def mark_final_attempt(self):
        """
            Mark the wait context as being in the final attempt condition.
        """
        self._final_attempt = True
        return

    def mark_time(self):
        """
            Called to mark the current time in the :class:`WaitContext` instance.
        """
        self._now_time = datetime.now()
        return

    def mark_timeout(self):
        """
            Called to mark the wait context as timed out.
        """
        self._wait_state = TimeoutState.TimedOut
        return

    def reduce_delay(self, secs):
        """
            Reduce the wait start delay.
        """
        if secs > self._delay:
            self._delay = 0
        else:
            self._delay = self._delay - secs
        return

    def should_continue(self) -> bool:
        """
            Indicates if a wait condition should continue based on time specifications and context.
        """
        self._now_time = datetime.now()
        self._loop_counter += 1

        scont = True

        if self._wait_state == TimeoutState.Completed:
            scont = False
        elif self._now_time > self._end_time:
            scont = False

        return scont



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class MediaManagement2ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:MediaManagement:2' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'MediaManagement2'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "MediaSessionInfo": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetMediaCapabilities(self, TSMediaCapabilityInfo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaCapabilities action.

            :returns: "SupportedMediaCapabilityInfo"
        """
        arguments = {
            "TSMediaCapabilityInfo": TSMediaCapabilityInfo,
        }

        out_params = self.call_action("GetMediaCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedMediaCapabilityInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaSessionInfo(self, TargetMediaSessionID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaSessionInfo action.

            :returns: "MediaSessionInfoList"
        """
        arguments = {
            "TargetMediaSessionID": TargetMediaSessionID,
        }

        out_params = self.call_action("GetMediaSessionInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MediaSessionInfoList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ModifyMediaSession(self, TargetMediaSessionID, NewMediaCapabilityInfo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ModifyMediaSession action.

            :returns: "TCMediaCapabilityInfo"
        """
        arguments = {
            "TargetMediaSessionID": TargetMediaSessionID,
            "NewMediaCapabilityInfo": NewMediaCapabilityInfo,
        }

        out_params = self.call_action("ModifyMediaSession", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TCMediaCapabilityInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_StartMediaSession(self, TSMediaCapabilityInfo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartMediaSession action.

            :returns: "MediaSessionID", "TCMediaCapabilityInfo"
        """
        arguments = {
            "TSMediaCapabilityInfo": TSMediaCapabilityInfo,
        }

        out_params = self.call_action("StartMediaSession", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MediaSessionID", "TCMediaCapabilityInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_StopMediaSession(self, TargetMediaSessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StopMediaSession action.
        """
        arguments = {
            "TargetMediaSessionID": TargetMediaSessionID,
        }

        self.call_action("StopMediaSession", arguments=arguments, aspects=aspects)

        return



"""
.. module:: compat
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains functions that are utilized to provide compatibility across different
               operating system platforms and python version.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Union

import traceback

import sys
import importlib

from types import ModuleType

NoneType = type(None)

is_python3 = sys.version_info[0] == 3
is_python_pre_3_5 = (is_python3 and sys.version_info[1] < 5)

def import_by_name(modulename: str) -> ModuleType:
    """
        Imports a module by name.
    """

    mod = None
    if modulename in sys.modules:
        mod = sys.modules[modulename]
    else:
        mod = importlib.import_module(modulename)

    return mod

def import_file(name: str, loc: str, by_file_only=False) -> ModuleType:
    """
        Import module from a file. Used to load models from a directory.

        :param unicode name: Name of module to load.
        :param (unicode / Path) loc: Path to the file.

        returns: The loaded module.
    """
    mod = None
    if name in sys.modules:
        mod = sys.modules[name]

    if mod is None:
        if is_python_pre_3_5:
            import imp # pylint: disable=import-outside-toplevel

            mod = imp.load_source(name, loc)
        else:

            while True:
                if not by_file_only:
                    # First try to import the module by name only
                    try:
                        mod = importlib.import_module(name)
                        break
                    except ImportError:
                        errmsg = traceback.format_exc()
                        pass

                spec = importlib.util.spec_from_file_location(name, str(loc))
                mod = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(mod)
                sys.modules[name] = mod
                break

    return mod

def bytes_cast(val: Union[str, bytes]) -> bytes:
    """
        Ensures that the 'val' parameter is a bytes object.

        :param val: A str or bytes value that needs to be converted to bytes.
        :param val: str or bytes

        :returns: The input value converted to a bytes
    """
    if isinstance(val, str):
        val = val.encode('utf-8')
    return val

def str_cast(val: Union[str, bytes]) -> str:
    """
        Ensures that the 'val' parameter is a str object.

        :param val: A str or bytes value that needs to be converted to str.
        :param val: str or bytes

        :returns: The input value converted to a str
    """
    if isinstance(val, bytes):
        val = val.decode('utf-8')
    return val



"""
.. module:: dnsquestion
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the DnsQuestion object which represents a DNS question to be posed to the DNS service.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import struct

from akit.interop.dns.dnsconst import DnsRecordClass, DnsRecordType
from akit.interop.dns.dnsrecord import DnsRecord

class DnsQuestion:
    """
        The :class:`DnsQuestion` object is utilized to formulate DNS querys and to check to
        see if a :class:`DnsRecord` answers the query provided.
    """
    def __init__(self, name: str, rtype: DnsRecordType, rclass: DnsRecordClass) -> None:
        self._name = name
        self._rtype = rtype
        self._rclass = DnsRecordClass(rclass & DnsRecordClass.MASK)
        self._unique = (rclass & DnsRecordClass.UNIQUE) != 0
        return

    @property
    def name(self):
        return self._name

    @property
    def rclass(self):
        return self._rclass

    @property
    def rtype(self):
        return self._rtype

    @property
    def unique(self):
        return self._unique

    def answered_by(self, rec: DnsRecord) -> bool:
        """
            Returns true if the question is answered by the record.
        """
        bval = False
        if self._rclass == rec._rclass:
            if self._rtype == rec._rtype or self._rtype == DnsRecordType.ANY:
               bval = self._name == rec._name
        return bval

    def as_query_bytes(self) -> str:
        """
            Formats the DNS question as a byte stream for transmission.
        """

        brtype = self._rtype.name.encode("utf-8")
        brclass = self._rclass.name.encode("utf-8")

        name_len = len(self._name)

        dnsstr = b"%s/x00%s,%s" % ('question', brtype, brclass)

        if self._unique:
            dnsstr += "-unique,"
        else:
            dnsstr += ","

        dnsstr += self._name

        dnsbytes = dnsstr.encode("utf-8")

        return dnsbytes



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANCommonInterfaceConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANCommonInterfaceConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "ActiveConnectionDeviceContainer": { "data_type": "string", "default": None, "allowed_list": None},
        "ActiveConnectionServiceID": { "data_type": "string", "default": None, "allowed_list": None},
        "EnabledForInternet": { "data_type": "boolean", "default": None, "allowed_list": None},
        "Layer1DownstreamMaxBitRate": { "data_type": "ui4", "default": None, "allowed_list": None},
        "Layer1UpstreamMaxBitRate": { "data_type": "ui4", "default": None, "allowed_list": None},
        "MaximumActiveConnections": { "data_type": "ui2", "default": None, "allowed_list": None},
        "NumberOfActiveConnections": { "data_type": "ui2", "default": None, "allowed_list": None},
        "PhysicalLinkStatus": { "data_type": "string", "default": None, "allowed_list": "['Up', 'Down']"},
        "TotalBytesReceived": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalBytesSent": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalPacketsReceived": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalPacketsSent": { "data_type": "ui4", "default": None, "allowed_list": None},
        "WANAccessProvider": { "data_type": "string", "default": None, "allowed_list": None},
        "WANAccessType": { "data_type": "string", "default": None, "allowed_list": "['DSL', 'POTS', 'Cable', 'Ethernet']"},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetActiveConnection(self, NewActiveConnectionIndex, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetActiveConnection action.

            :returns: "NewActiveConnDeviceContainer", "NewActiveConnectionServiceID"
        """
        arguments = {
            "NewActiveConnectionIndex": NewActiveConnectionIndex,
        }

        out_params = self.call_action("GetActiveConnection", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewActiveConnDeviceContainer", "NewActiveConnectionServiceID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCommonLinkProperties(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCommonLinkProperties action.

            :returns: "NewWANAccessType", "NewLayer1UpstreamMaxBitRate", "NewLayer1DownstreamMaxBitRate", "NewPhysicalLinkStatus"
        """
        arguments = { }

        out_params = self.call_action("GetCommonLinkProperties", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewWANAccessType", "NewLayer1UpstreamMaxBitRate", "NewLayer1DownstreamMaxBitRate", "NewPhysicalLinkStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetEnabledForInternet(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetEnabledForInternet action.

            :returns: "NewEnabledForInternet"
        """
        arguments = { }

        out_params = self.call_action("GetEnabledForInternet", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewEnabledForInternet",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMaximumActiveConnections(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMaximumActiveConnections action.

            :returns: "NewMaximumActiveConnections"
        """
        arguments = { }

        out_params = self.call_action("GetMaximumActiveConnections", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewMaximumActiveConnections",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTotalBytesReceived(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTotalBytesReceived action.

            :returns: "NewTotalBytesReceived"
        """
        arguments = { }

        out_params = self.call_action("GetTotalBytesReceived", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalBytesReceived",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTotalBytesSent(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTotalBytesSent action.

            :returns: "NewTotalBytesSent"
        """
        arguments = { }

        out_params = self.call_action("GetTotalBytesSent", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalBytesSent",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTotalPacketsReceived(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTotalPacketsReceived action.

            :returns: "NewTotalPacketsReceived"
        """
        arguments = { }

        out_params = self.call_action("GetTotalPacketsReceived", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalPacketsReceived",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTotalPacketsSent(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTotalPacketsSent action.

            :returns: "NewTotalPacketsSent"
        """
        arguments = { }

        out_params = self.call_action("GetTotalPacketsSent", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalPacketsSent",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetWANAccessProvider(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetWANAccessProvider action.

            :returns: "NewWANAccessProvider"
        """
        arguments = { }

        out_params = self.call_action("GetWANAccessProvider", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewWANAccessProvider",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetEnabledForInternet(self, NewEnabledForInternet, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetEnabledForInternet action.
        """
        arguments = {
            "NewEnabledForInternet": NewEnabledForInternet,
        }

        self.call_action("SetEnabledForInternet", arguments=arguments, aspects=aspects)

        return



"""
.. module:: landscapeconfigurationlayer
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the Landscape related classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

# ====================================================================================
#
#                               OPERATIONAL LAYER
#
# ====================================================================================

from typing import List, Optional, Union

import inspect
import json
import os
import threading

import zeroconf

from akit.exceptions import AKitConfigurationError, AKitSemanticError
from akit.friendlyidentifier import FriendlyIdentifier
from akit.interop.landscaping.landscapedevice import LandscapeDevice

from akit.paths import get_path_for_output

from akit.interop.coordinators.powercoordinator import PowerCoordinator
from akit.interop.coordinators.serialcoordinator import SerialCoordinator

from akit.interop.landscaping.landscapedevice import LandscapeDevice

from akit.interop.dns.mdnsservicecatalog import MdnsServiceCatalog

class LandscapeOperationalLayer:
    """
    """

    _operational_gate = None

    MDNS_BROWSE_TYPES = ["_http._tcp.local.", "_sonos._tcp.local."]

    def __init__(self):
        self._power_coord = None
        self._serial_coord = None

        self._upnp_coord = None
        self._ssh_coord = None

        self._active_devices = {}

        self._device_pool = {}
        self._checked_out_devices = {}

        self._activation_errors = []

        self._first_contact_results = None

        self._integration_points_activated = {}
        self._integration_point_activation_counter = 0

        super().__init__()

        self._zeroconf = zeroconf.Zeroconf()
        self._zeroconf_catalog = MdnsServiceCatalog(self.logger)
        self._zeroconf_browser = zeroconf.ServiceBrowser(self._zeroconf, self.MDNS_BROWSE_TYPES, self._zeroconf_catalog)

        return

    @property
    def ssh_coord(self):
        """
            Returns a the :class:`SshPoolCoordinator` that is used to manage ssh devices.
        """
        self._ensure_activation()
        return self._ssh_coord

    @property
    def upnp_coord(self):
        """
            Returns a the :class:`UpnpCoordinator` that is used to manage upnp devices.
        """
        self._ensure_activation()
        return self._upnp_coord

    def activate_integration_point(self, role: str, coordinator_constructor: callable):
        """
            This method should be called from the attach_to_environment methods from individual couplings
            in order to register the base level integrations.  Integrations can be hierarchical so it
            is only necessary to register the root level integration couplings, the descendant couplings can
            be called from the root level couplings.

            :param role: The name of a role to assign for a coupling.
            :param coupling: The coupling to register for the associated role.
        """

        if role.startswith("coordinator/"):
            
            if "coordinator/serial" not in self._integration_points_activated:
                self._integration_points_activated["coordinator/serial"] = True

            if "coordinator/power" not in self._integration_points_activated:
                self._integration_points_activated["coordinator/power"] = True

            _, coord_type = role.split("/")
            if coord_type == "upnp" or coord_type == "ssh":
                if role not in self._integration_points_activated:
                    self._integration_points_activated[role] = coordinator_constructor
                else:
                    raise AKitSemanticError("Attempted to activate the UPNP coordinator twice.") from None
            else:
                raise AKitSemanticError("Unknown coordinator type '%s'." % role) from None
        else:
            raise AKitSemanticError("Don't know how to activate integration point of type '%s'." % role) from None

        return

    def activate_operations(self, allow_missing_devices: bool=False, allow_unknown_devices: bool=False, upnp_recording: bool=False,
                            validate_features: bool=True, validate_topology: bool=True):

        thisType = type(self)

        self.landscape_lock.acquire()
        try:

            if thisType._operational_gate is None:
                thisType._operational_gate = threading.Event()
                thisType._operational_gate.clear()

                # Don't hold the landscape like while we wait for the
                # landscape to be activated
                self.landscape_lock.release()
                try:
                    if "coordinator/serial" in self._integration_points_activated:
                        self._activate_serial_coordinator()
                    
                    if "coordinator/power" in self._integration_points_activated:
                        self._activate_power_coordinator()
                    
                    if "coordinator/upnp" in self._integration_points_activated:
                        coordinator_constructor = self._integration_points_activated["coordinator/upnp"]
                        self._activate_upnp_coordinator(coordinator_constructor)
                    
                    if "coordinator/ssh" in self._integration_points_activated:
                        coordinator_constructor = self._integration_points_activated["coordinator/ssh"]
                        self._activate_ssh_coordinator(coordinator_constructor)

                    self._establish_connectivity(allow_missing_devices=allow_missing_devices, 
                                                 allow_unknown_devices=allow_unknown_devices,
                                                 upnp_recording=upnp_recording)

                    self._zeroconf_browser.cancel()

                    if validate_features:
                        self._features_validate()

                    if validate_topology:
                        self._topology_validate()

                    self._operational_gate.set()

                finally:
                    self.landscape_lock.acquire()

            else:

                # Don't hold the landscape like while we wait for the
                # landscape to be activated
                self.landscape_lock.release()
                try:
                    # Because the landscape is a global singleton and because
                    # we were not the first thread to call the activate method,
                    # wait for the first calling thread to finish activating the
                    # Landscape before we return allowing other use of the Landscape
                    # singleton
                    self._operational_gate.wait()
                finally:
                    self.landscape_lock.acquire()

        finally:
            self.landscape_lock.release()

        return

    def checkin_device(self, device: LandscapeDevice):
        """
            Returns a landscape device to the the available device pool.
        """
        self._ensure_activation()

        identity = device.identity

        if identity not in self._checked_out_devices:
            errmsg = "Attempting to checkin a device that is not checked out. {}".format(device)
            raise AKitSemanticError(errmsg)

        self.landscape_lock.acquire()
        try:
            self._device_pool[identity] = device
            del self._checked_out_devices[identity]
        finally:
            self.landscape_lock.release()

        return
    
    def checkin_multiple_devices(self, devices: List[LandscapeDevice]):
        """
            Returns a landscape device to the the available device pool.
        """
        self._ensure_activation()

        checkin_errors = []

        self.landscape_lock.acquire()
        try:

            for dev in devices:
                identity = dev.identity

                if identity not in self._checked_out_devices:
                    self._device_pool[identity] = dev
                    checkin_errors.append(dev)

            if len(checkin_errors) > 0:
                err_msg_lines = [
                    "Attempting to checkin a device that is not checked out.",
                    "DEVICES:"
                ]
                for dev in checkin_errors:
                    err_msg_lines.append("    {}".format(dev))

                err_msg = os.linesep.join(err_msg_lines)
                raise AKitSemanticError(err_msg)

            for dev in devices:
                identity = dev.identity

                if identity in self._checked_out_devices:
                    self._device_pool[identity] = dev
                    del self._checked_out_devices[identity]

        finally:
            self.landscape_lock.release()

        return

    def checkout_a_device_by_modelName(self, modelName: str) -> Optional[LandscapeDevice]:
        """
            Checks out a single device from the available pool using the modelName match
            criteria provided.
        """
        self._ensure_activation()

        device = None

        device_list = self.checkout_devices_by_match("modelName", modelName, count=1)
        if len(device_list) > 0:
            device = device_list[0]

        return device

    def checkout_a_device_by_modelNumber(self, modelNumber: str) -> Optional[LandscapeDevice]:
        """
            Checks out a single device from the available pool using the modelNumber match
            criteria provided.
        """
        self._ensure_activation()

        device = None

        device_list = self.checkout_devices_by_match("modelNumber", modelNumber, count=1)
        if len(device_list) > 0:
            device = device_list[0]

        return device

    def checkout_device(self, device: LandscapeDevice):
        """
            Checks out the specified device from the device pool.
        """
        self._ensure_activation()

        self.landscape_lock.acquire()
        try:
            self._locked_checkout_device(device)
        finally:
            self.landscape_lock.release()

        return
    
    def checkout_device_list(self, device_list: List[LandscapeDevice]):
        """
            Checks out the list of specified devices from the device pool.
        """
        self._ensure_activation()

        self.landscape_lock.acquire()
        try:
            for device in device_list:
                self._locked_checkout_device(device)
        finally:
            self.landscape_lock.release()

        return

    def checkout_devices_by_match(self, match_type: str, *match_params, count=None) -> List[LandscapeDevice]:
        """
            Checks out the devices that are found to correspond with the match criteria provided.  If the
            'count' parameter is passed, then the number of devices that are checked out is limited to
            count matching devices.
        """
        self._ensure_activation()

        match_list = None

        self.landscape_lock.acquire()
        try:
            match_list = self.list_available_devices_by_match(match_type, *match_params, count=count)

            for device in match_list:
                self._locked_checkout_device(device)
        finally:
            self.landscape_lock.release()

        return match_list

    def checkout_devices_by_modelName(self, modelName:str , count=None) -> List[LandscapeDevice]:
        """
            Checks out the devices that are found to correspond with the modelName match criteria provided.
            If the 'count' parameter is passed, the the number of devices that are checked out is limited to
            count matching devices.
        """
        self._ensure_activation()

        device_list = self.checkout_devices_by_match("modelName", modelName, count=count)

        return device_list


    def checkout_devices_by_modelNumber(self, modelNumber: str, count=None) -> List[LandscapeDevice]:
        """
            Checks out the devices that are found to correspond with the modelNumber match criteria provided.
            If the 'count' parameter is passed, the the number of devices that are checked out is limited to
            count matching devices.
        """
        self._ensure_activation()

        device_list = self.checkout_devices_by_match("modelNumber", modelNumber, count=count)

        return device_list

    def diagnostic(self, diaglabel: str, diags: dict):
        """
            Can be called in order to perform a diagnostic capture across the test landscape.

            :param diaglabel: The label to use for the diagnostic.
            :param diags: A dictionary of diagnostics to run.
        """
        self._ensure_activation()

        return

    def first_contact(self) -> List[str]:
        """
            The `first_contact` method provides a mechanism for the verification of connectivity with
            enterprise resources that is seperate from the initial call to `establish_connectivity`.

            :returns list: list of failing entities
        """
        error_list = []
        return error_list

    def list_available_devices(self) -> List[LandscapeDevice]:
        """
            Returns the list of devices from the landscape device pool.  This will
            skip any device that has a "skip": true member.
        """
        self._ensure_activation()

        device_list = None

        self.landscape_lock.acquire()
        try:
            device_list = [dev for dev in self._device_pool.values()]
        finally:
            self.landscape_lock.release()

        return device_list

    def list_available_devices_by_match(self, match_type, *match_params, count=None) -> List[LandscapeDevice]:
        """
            Creates and returns a list of devices from the available devices pool that are found
            to correspond to the match criteria provided.  If a 'count' parameter is passed
            then the number of devices returned is limited to count devices.

            .. note:: This API does not perform a checkout of the devices returns so the
                      caller should not consider themselves to the the owner of the devices.
        """
        matching_devices = []
        device_list = self.list_available_devices()

        for dev in device_list:
            if dev.match_using_params(match_type, *match_params):
                matching_devices.append(dev)
                if count is not None and len(matching_devices) >= count:
                    break

        return matching_devices

    def list_devices_by_match(self, match_type, *match_params, count=None) -> List[LandscapeDevice]:
        """
            Creates and returns a list of devices that are found to correspond to the match
            criteria provided.  If a 'count' parameter is passed then the number of devices
            returned is limited to count devices.
        """
        matching_devices = []
        device_list = self.get_devices()

        for dev in device_list:
            if dev.match_using_params(match_type, *match_params):
                matching_devices.append(dev)
                if count is not None and len(matching_devices) >= count:
                    break

        return matching_devices

    def list_devices_by_modelName(self, modelName, count=None) -> List[LandscapeDevice]:
        """
            Creates and returns a list of devices that are found to correspond to the modelName
            match criteria provided.  If a 'count' parameter is passed then the number of devices
            returned is limited to count devices.
        """

        matching_devices = self.list_devices_by_match("modelName", modelName, count=count)

        return matching_devices

    def list_devices_by_modelNumber(self, modelNumber, count=None) -> List[LandscapeDevice]:
        """
            Creates and returns a list of devices that are found to correspond to the modelNumber
            match criteria provided.  If a 'count' parameter is passed then the number of devices
            returned is limited to count devices.
        """

        matching_devices = self.list_devices_by_match("modelNumber", modelNumber, count=count)

        return matching_devices

    def list_devices_from_hint_list(self, hint_list: List[str]) -> List[LandscapeDevice]:
        """
            Creates and returns a list of devices that are found that are matched from the list
            of hints provided.
        """
        matching_devices = []
        device_list = self.get_devices()

        remaining_hints = [hint for hint in hint_list]
        while len(remaining_hints) > 0:
            hint = remaining_hints.pop()

            for dev in device_list:
                if dev.friendly_id.match(hint):
                    matching_devices.append(dev)

        return matching_devices

    def lookup_credential(self, credential_name) -> Union[str, None]:
        """
            Looks up a credential.
        """
        cred_info = None
        
        if credential_name in self._credentials:
            cred_info = self._credentials[credential_name]

        return cred_info

    def lookup_device_by_identity(self, identity: Union[str, FriendlyIdentifier]) -> Optional[LandscapeDevice]:
        """
            Looks up a single device that is found to correspond to the identity.
        """
        found_device = None

        device_list = self.get_devices()
        for device in device_list:
            if device.friendly_id.match(identity):
                found_device = device
                break

        return found_device

    def lookup_device_by_modelName(self, modelName) -> Optional[LandscapeDevice]:
        """
            Looks up a single device that is found to correspond to the modelName match criteria
            provided.
        """
        found_device = None

        matching_devices = self.list_devices_by_match("modelName", modelName, count=1)
        if len(matching_devices) > 0:
            found_device = matching_devices[0]

        return found_device

    def lookup_device_by_modelNumber(self, modelNumber) -> Optional[LandscapeDevice]:
        """
            Looks up a single device that is found to correspond to the modelNumber match criteria
            provided.
        """
        found_device = None

        matching_devices = self.list_devices_by_match("modelNumber", modelNumber, count=1)
        if len(matching_devices) > 0:
            found_device = matching_devices[0]

        return found_device

    def lookup_power_agent(self, power_mapping: dict) -> Union[dict, None]:
        """
            Looks up a power agent by name.
        """
        power_agent = self._power_coord.lookup_agent(power_mapping)
        return power_agent

    def lookup_serial_agent(self, serial_mapping: str) -> Union[dict, None]:
        """
            Looks up a serial agent name.
        """
        serial_agent = self._serial_coordinator.lookup_agent(serial_mapping)
        return serial_agent

    def _activate_power_coordinator(self):
        """
            Initializes the power coordinator according the the information specified in the
            'power' portion of the configuration file.
        """
        pod_info = self._landscape_info["pod"]

        # We need to initialize the power before attempting to initialize any devices, so the
        # devices will be able to lookup serial connections as they are initialized
        if "power" in pod_info:
            coord_config = pod_info["power"]
            self._power_coord = PowerCoordinator(self, coord_config=coord_config)

        return

    def _activate_serial_coordinator(self):
        """
            Initializes the serial coordinator according the the information specified in the
            'serial' portion of the configuration file.
        """
        pod_info = self._landscape_info["pod"]

        # We need to initialize the serial before attempting to initialize any devices, so the
        # devices will be able to lookup serial connections as they are initialized
        if "serial" in pod_info:
            coord_config = pod_info["serial"]
            self._serial_coord = SerialCoordinator(self, coord_config=coord_config)

        return

    def _activate_ssh_coordinator(self, coordinator_constructor):
        """
            Initializes the ssh coordinator according the the information specified in the
            'devices' portion of the configuration file.
        """
        self._has_ssh_devices = True
        self._ssh_coord = coordinator_constructor(self)

        return

    def _activate_upnp_coordinator(self, coordinator_constructor):
        """
            Initializes the upnp coordinator according the the information specified in the
            'devices' portion of the configuration file.
        """

        self._has_upnp_devices = True        
        self._upnp_coord = coordinator_constructor(self)

        return

    def _ensure_activation(self):
        """
            Called by methods that require Landscape activation in order to make sure the 'activate' method
            has been called before the attempted use of the specified method.

            :param method: The name of the method guarding against the use of a Landscape that has not been
                           activated.
        """
        if self._operational_gate is not None:
            self._operational_gate.wait()
        else:
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            guarded_method = calframe[1][3]

            errmsg = "The Landscape must be activated before calling the '%s' method." % guarded_method
            raise AKitSemanticError(errmsg) from None

        return
    
    def _establish_connectivity(self, allow_missing_devices: bool = False, allow_unknown_devices: bool = False, upnp_recording: bool = False) -> List[str]:
        """
            The `_establish_connectivity` method provides a mechanism for the verification of connectivity with
            enterprise resources.

            :returns list: list of failing entities
        """

        error_list = []
        connectivity_results = {}

        if self._has_upnp_devices:
            integration_cls = self._integration_points_registered["coordinator/upnp"]
            upnp_error_list, upnp_connectivity_results = integration_cls.establish_connectivity(allow_missing_devices=allow_missing_devices,
                upnp_recording=upnp_recording, allow_unknown_devices=allow_unknown_devices)
            error_list.extend(upnp_error_list)
            connectivity_results.update(upnp_connectivity_results)

        if self._has_ssh_devices:
            integration_cls = self._integration_points_registered["coordinator/ssh"]
            ssh_error_list, ssh_connectivity_results = integration_cls.establish_connectivity(allow_missing_devices=allow_missing_devices)
            error_list.extend(ssh_error_list)
            connectivity_results.update(ssh_connectivity_results)

        self._first_contact_results = connectivity_results

        self._log_scan_results(connectivity_results, )

        return error_list

    def _features_validate(self):
        """
            Validates the device features specified in the landscape configuration file.
        """
        return

    def _internal_activate_device(self, identity):
        """
            Activates a device by copying a reference to the device from the all_devices
            pool to the active_devices and device_pool tables to make the device available
            for active use.
        """
        errmsg = None

        self.landscape_lock.acquire()
        try:
            device = None

            # Add the device to all devices, all devices does not change
            # based on check-out or check-in activity
            if identity in self._all_devices:
                device = self._all_devices[identity]

            if device is not None:
                # Add the device to the device pool, the device pool is used
                # for tracking device availability for check-out
                self._active_devices[identity] = device
                self._device_pool[identity] = device
            else:
                errmsg = "Attempt made to activate an unknown device. identity=%s" % identity

        finally:
            self.landscape_lock.release()

        return errmsg

    def _internal_get_upnp_coord(self):
        """
            Internal method to get a reference to the upnp coordinator.  This provides access
            to the upnp coordinator reference in the middle of activation and bypasses normal
            activation thread synchronization mechanisms.  It should only be used after the upnp
            coordinator has been activated.
        """
        return self._upnp_coord

    def _intenal_scan_activated_devices_for_power(self) -> bool:
        """
            Go through all of the activated device types such as SSH and
            UPNP look for power automation requirements.
        """
        return

    def _intenal_scan_activated_devices_for_serial(self) -> bool:
        """
            Go through all of the activated device types such as SSH and
            UPNP look for power automation requirements.
        """
        return

    def _locked_checkout_device(self, device) -> Optional[LandscapeDevice]:

        rtn_device = None

        identity = device.identity
        if identity not in self._device_pool:
            raise AKitSemanticError("A device is being checked out, that is not in the device pool.") from None

        rtn_device = self._device_pool[identity]

        del self._device_pool[identity]
        self._checked_out_devices[identity] = rtn_device

        return rtn_device

    def _log_device_activation_results(self):

        landscape_first_contact_result_file = os.path.join(get_path_for_output(), "landscape-first-contact-results.json")
        with open(landscape_first_contact_result_file, 'w') as fcrf:
            json.dump(self._first_contact_results, fcrf, indent=4)

        if len(self._activation_errors) > 0:
            errmsg_lines = [
                "Encountered device activation errors.",
                "ACTIVATION ERROR LIST:"
            ]
            for aerror in self._activation_errors:
                errmsg_lines.append("    %s" % aerror)

            errmsg = os.linesep.join(errmsg_lines)
            raise AKitConfigurationError(errmsg) from None

        return
    
    def _log_scan_results(self, scan_results: dict,):
        """
            Logs the results of the device scan.
            :param scan_results: A combined dictionary of scan results.
        """
        log_landscape_scan = self.context.lookup("/environment/behaviors/log-landscape-scan")
        if log_landscape_scan:

            landscape_scan_result_file = os.path.join(get_path_for_output(), "landscape-startup-scan.json")
            with open(landscape_scan_result_file, 'w') as srf:
                json.dump(scan_results, srf, indent=4)

        return

    def _topology_validate(self):
        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import click

from akit.cli.cmdtree.testing.jobs import group_akit_testing_jobs

from akit.cli.cmdtree.testing.query import command_akit_testing_query
from akit.cli.cmdtree.testing.run import command_akit_testing_run

@click.group("testing", help="Contains command for working with tests and test results.")
def group_akit_testing():
    return

group_akit_testing.add_command(group_akit_testing_jobs)

group_akit_testing.add_command(command_akit_testing_query)
group_akit_testing.add_command(command_akit_testing_run)



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class SoftwareManagement1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:SoftwareManagement:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'SoftwareManagement1'

    SERVICE_DEFAULT_VARIABLES = {
        "ActiveEUIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "DUIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "EUIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "ErrorEUIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "OperationIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "RunningEUIDs": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetActiveEUIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetActiveEUIDs action.

            :returns: "ActiveEUIDs"
        """
        arguments = { }

        out_params = self.call_action("GetActiveEUIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ActiveEUIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDUIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDUIDs action.

            :returns: "DUIDs"
        """
        arguments = { }

        out_params = self.call_action("GetDUIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DUIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDUInfo(self, DUID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDUInfo action.

            :returns: "DUName", "DUVersion", "DUType", "DUState", "DUURI"
        """
        arguments = {
            "DUID": DUID,
        }

        out_params = self.call_action("GetDUInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DUName", "DUVersion", "DUType", "DUState", "DUURI",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetEUIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetEUIDs action.

            :returns: "EUIDs"
        """
        arguments = { }

        out_params = self.call_action("GetEUIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("EUIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetEUInfo(self, EUID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetEUInfo action.

            :returns: "EUName", "EUVersion", "EURequestedState", "EUExecutionState"
        """
        arguments = {
            "EUID": EUID,
        }

        out_params = self.call_action("GetEUInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("EUName", "EUVersion", "EURequestedState", "EUExecutionState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetErrorEUIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetErrorEUIDs action.

            :returns: "ErrorEUIDs"
        """
        arguments = { }

        out_params = self.call_action("GetErrorEUIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ErrorEUIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOperationIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOperationIDs action.

            :returns: "OperationIDs"
        """
        arguments = { }

        out_params = self.call_action("GetOperationIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOperationInfo(self, OperationID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOperationInfo action.

            :returns: "OperationState", "TargetedIDs", "Action", "ErrorDescription", "AdditionalInfo"
        """
        arguments = {
            "OperationID": OperationID,
        }

        out_params = self.call_action("GetOperationInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationState", "TargetedIDs", "Action", "ErrorDescription", "AdditionalInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRunningEUIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRunningEUIDs action.

            :returns: "RunningEUIDs"
        """
        arguments = { }

        out_params = self.call_action("GetRunningEUIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RunningEUIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Install(self, DUURI, DUType, HandleDependencies, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Install action.

            :returns: "OperationID"
        """
        arguments = {
            "DUURI": DUURI,
            "DUType": DUType,
            "HandleDependencies": HandleDependencies,
        }

        out_params = self.call_action("Install", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Start(self, EUID, HandleDependencies, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Start action.

            :returns: "OperationID"
        """
        arguments = {
            "EUID": EUID,
            "HandleDependencies": HandleDependencies,
        }

        out_params = self.call_action("Start", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Stop(self, EUID, HandleDependencies, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Stop action.

            :returns: "OperationID"
        """
        arguments = {
            "EUID": EUID,
            "HandleDependencies": HandleDependencies,
        }

        out_params = self.call_action("Stop", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Uninstall(self, DUID, HandleDependencies, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Uninstall action.

            :returns: "OperationID"
        """
        arguments = {
            "DUID": DUID,
            "HandleDependencies": HandleDependencies,
        }

        out_params = self.call_action("Uninstall", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Update(self, DUID, NewDUURI, HandleDependencies, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Update action.

            :returns: "OperationID"
        """
        arguments = {
            "DUID": DUID,
            "NewDUURI": NewDUURI,
            "HandleDependencies": HandleDependencies,
        }

        out_params = self.call_action("Update", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OperationID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class DigitalSecurityCameraSettings1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:DigitalSecurityCameraSettings:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'DigitalSecurityCameraSettings1'

    SERVICE_DEFAULT_VARIABLES = {
        "AvailableRotations": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "AutomaticWhiteBalance": { "data_type": "boolean", "default": "1", "allowed_list": None},
        "Brightness": { "data_type": "ui1", "default": "50", "allowed_list": None},
        "ColorSaturation": { "data_type": "ui1", "default": "50", "allowed_list": None},
        "DefaultRotation": { "data_type": "string", "default": None, "allowed_list": None},
        "FixedWhiteBalance": { "data_type": "ui4", "default": "3000", "allowed_list": None},
    }

    def action_DecreaseBrightness(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DecreaseBrightness action.
        """
        arguments = { }

        self.call_action("DecreaseBrightness", arguments=arguments, aspects=aspects)

        return

    def action_DecreaseColorSaturation(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DecreaseColorSaturation action.
        """
        arguments = { }

        self.call_action("DecreaseColorSaturation", arguments=arguments, aspects=aspects)

        return

    def action_GetAutomaticWhiteBalance(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutomaticWhiteBalance action.

            :returns: "RetAutomaticWhiteBalance"
        """
        arguments = { }

        out_params = self.call_action("GetAutomaticWhiteBalance", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetAutomaticWhiteBalance",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAvailableRotations(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAvailableRotations action.

            :returns: "RetAvailableRotations"
        """
        arguments = { }

        out_params = self.call_action("GetAvailableRotations", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetAvailableRotations",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBrightness(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBrightness action.

            :returns: "RetBrightness"
        """
        arguments = { }

        out_params = self.call_action("GetBrightness", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetBrightness",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetColorSaturation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetColorSaturation action.

            :returns: "RetColorSaturation"
        """
        arguments = { }

        out_params = self.call_action("GetColorSaturation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetColorSaturation",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDefaultRotation(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDefaultRotation action.

            :returns: "RetRotation"
        """
        arguments = { }

        out_params = self.call_action("GetDefaultRotation", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetRotation",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFixedWhiteBalance(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFixedWhiteBalance action.

            :returns: "RetFixedWhiteBalance"
        """
        arguments = { }

        out_params = self.call_action("GetFixedWhiteBalance", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetFixedWhiteBalance",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_IncreaseBrightness(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the IncreaseBrightness action.
        """
        arguments = { }

        self.call_action("IncreaseBrightness", arguments=arguments, aspects=aspects)

        return

    def action_IncreaseColorSaturation(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the IncreaseColorSaturation action.
        """
        arguments = { }

        self.call_action("IncreaseColorSaturation", arguments=arguments, aspects=aspects)

        return

    def action_SetAutomaticWhiteBalance(self, NewAutomaticWhiteBalance, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutomaticWhiteBalance action.
        """
        arguments = {
            "NewAutomaticWhiteBalance": NewAutomaticWhiteBalance,
        }

        self.call_action("SetAutomaticWhiteBalance", arguments=arguments, aspects=aspects)

        return

    def action_SetBrightness(self, NewBrightness, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBrightness action.
        """
        arguments = {
            "NewBrightness": NewBrightness,
        }

        self.call_action("SetBrightness", arguments=arguments, aspects=aspects)

        return

    def action_SetColorSaturation(self, NewColorSaturation, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetColorSaturation action.
        """
        arguments = {
            "NewColorSaturation": NewColorSaturation,
        }

        self.call_action("SetColorSaturation", arguments=arguments, aspects=aspects)

        return

    def action_SetDefaultRotation(self, NewRotation, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDefaultRotation action.
        """
        arguments = {
            "NewRotation": NewRotation,
        }

        self.call_action("SetDefaultRotation", arguments=arguments, aspects=aspects)

        return

    def action_SetFixedWhiteBalance(self, NewFixedWhiteBalance, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetFixedWhiteBalance action.
        """
        arguments = {
            "NewFixedWhiteBalance": NewFixedWhiteBalance,
        }

        self.call_action("SetFixedWhiteBalance", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class CloudProxyDevice1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:CloudProxyDevice:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'CloudProxyDevice1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {}



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Union, NamedTuple

class WifiChannelInfo(NamedTuple):
    channel: int
    frequency: int

_demo_channel_mapping = tuple(
  #  ch, freq
    (20, 2312),
    (21, 2337)
    (22, 2362)
)

DEMO_CHANNEL_MAPPING = tuple([WifiChannelInfo(ch, freq) for ch, freq in _demo_channel_mapping])

_g_channel_mapping = tuple(
  #  ch, freq
    ( 1, 2412),
    ( 2, 2417),
    ( 3, 2422),
    ( 4, 2427),
    ( 5, 2432),
    ( 6, 2437),
    ( 7, 2442),
    ( 8, 2447),
    ( 9, 2452),
    (10, 2457),
    (11, 2462),
    (12, 2467),
    (13, 2472),
    (14, 2477),
)

G_CHANNELS = tuple([ch for ch, _ in _g_channel_mapping])
G_FREQUENCIES = tuple([freq for _, freq in _g_channel_mapping])

G_CHANNEL_MAPPING = tuple([WifiChannelInfo(ch, freq) for ch, freq in _g_channel_mapping])

_n_channel_mapping = tuple(
  #  ch, freq
    (36, 5180),
    (40, 5200),
    (44, 5220),
    (48, 5240),
    
    (52, 5260),
    (56, 5280),
    (60, 5300),
    (64, 5320),
    
    (100, 5500),
    (104, 5520),
    (108, 5540),
    (112, 5560),
    (116, 5580),
    (120, 5600),
    (124, 5620),
    (128, 5640),
    (132, 5660),
    (136, 5680),
    (140, 5700),

    (149, 5745),
    (153, 5765),
    (157, 5785),
    (161, 5805),
    (165, 5825),
)

N_CHANNELS = tuple([ch for ch, _ in _n_channel_mapping])
N_FREQUENCIES = tuple([freq for _, freq in _n_channel_mapping])

N_CHANNEL_MAPPING = tuple([WifiChannelInfo(ch, freq) for ch, freq in _n_channel_mapping])

FULL_CHANNEL_MAPPING = (DEMO_CHANNEL_MAPPING + G_CHANNEL_MAPPING + N_CHANNEL_MAPPING)

def _find_channel_by(attr, value):
    result = filter(lambda wlan_chan: getattr(wlan_chan, attr) == int(value),
        FULL_CHANNEL_MAPPING)
    
    if not result:
        raise LookupError('Failed to find a Wifi Channel with <{}> = <{}>'.format(attr, value))
    
    channel = result[0]
    
    return channel

def lookup_channel_frequency(channel: int) -> int:
    """
        Looks up the frequency in MHz that is associated with a channel number

        :param channel:  the channel number

        :returns: the frequency associated with the channel
    """
    freq = _find_channel_by('channel', channel).frequency
    return freq

def lookup_channel_for_frequency(frequency: int) -> int:
    """
        Converts a channel frequency, in MHz, to its corresponding channel number

        :param frequency:  the channel frequency
        
        :returns: the channel number
    """
    channel = _find_channel_by('frequency', frequency).channel
    return channel

def is_2ghz_channel(channel: Union[int, str]) -> bool:
    """
        Check a channel number to see if its 2GHz or not

        :param channel: the channel number to check

        :returns: whether or not channel is a valid 2GHz channel number or frequency
    """
    rtnval = int(channel) in G_CHANNELS
    return rtnval

def is_2ghz_frequency(frequency: Union[int, str]) -> bool:
    """
        Check a frequency to see if it is associated with a 2GHz channel

        :param frequency: the frequency number to check

        :returns: whether or not frequency is associated with a valid 2GHz channel
    """
    rtnval = int(frequency) in G_FREQUENCIES
    return rtnval

def is_5ghz_channel(channel: Union[int, str]) -> bool:
    """
        Check a channel number to see if its 5GHz or not

        :param channel: the channel number to check

        :returns: whether or not channel is a valid 5GHz channel number or frequency
    """
    rtnval = int(channel) in N_CHANNELS
    return rtnval

def is_5ghz_frequency(frequency: Union[int, str]) -> bool:
    """
        Check a frequency to see if it is associated with a 5GHz channel

        :param frequency: the frequency number to check

        :returns: whether or not frequency is associated with a valid 5GHz channel
    """
    rtnval = int(frequency) in N_FREQUENCIES
    return rtnval



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class MusicServices1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'MusicServices1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:MusicServices:1'

    SERVICE_DEFAULT_VARIABLES = {
        "ServiceId": { "data_type": "ui4", "default": None, "allowed_list": None},
        "SessionId": { "data_type": "string", "default": None, "allowed_list": None},
        "Username": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "ServiceListVersion": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetSessionId(self, ServiceId, Username, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSessionId action.

            :returns: "SessionId"
        """
        arguments = {
            "ServiceId": ServiceId,
            "Username": Username,
        }

        out_params = self.call_action("GetSessionId", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SessionId",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ListAvailableServices(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ListAvailableServices action.

            :returns: "AvailableServiceDescriptorList", "AvailableServiceTypeList", "AvailableServiceListVersion"
        """
        arguments = { }

        out_params = self.call_action("ListAvailableServices", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AvailableServiceDescriptorList", "AvailableServiceTypeList", "AvailableServiceListVersion",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_UpdateAvailableServices(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdateAvailableServices action.
        """
        arguments = { }

        self.call_action("UpdateAvailableServices", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Calendar1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Calendar:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Calendar1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "CalendarItem": { "data_type": "string", "default": None, "allowed_list": None},
        "MemoInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "TriggeredItem": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddCalendarItems(self, Caltems, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddCalendarItems action.

            :returns: "ItemIDs"
        """
        arguments = {
            "Caltems": Caltems,
        }

        out_params = self.call_action("AddCalendarItems", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ItemIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeleteCalendarItems(self, ItemIDs, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteCalendarItems action.
        """
        arguments = {
            "ItemIDs": ItemIDs,
        }

        self.call_action("DeleteCalendarItems", arguments=arguments, aspects=aspects)

        return

    def action_GetCalendarItems(self, ItemIDs, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCalendarItems action.

            :returns: "Caltems"
        """
        arguments = {
            "ItemIDs": ItemIDs,
        }

        out_params = self.call_action("GetCalendarItems", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Caltems",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMemo(self, MemoID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMemo action.

            :returns: "MemoInfoList"
        """
        arguments = {
            "MemoID": MemoID,
        }

        out_params = self.call_action("GetMemo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MemoInfoList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTelCPNameList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTelCPNameList action.

            :returns: "TelCPName"
        """
        arguments = { }

        out_params = self.call_action("GetTelCPNameList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TelCPName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTriggeredItems(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTriggeredItems action.

            :returns: "TriggeredItemIDs"
        """
        arguments = { }

        out_params = self.call_action("GetTriggeredItems", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TriggeredItemIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_PostMemo(self, Memo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the PostMemo action.

            :returns: "MemoID"
        """
        arguments = {
            "Memo": Memo,
        }

        out_params = self.call_action("PostMemo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MemoID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RegisterItemDelivery(self, ItemDeliveryMethod, Expires, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterItemDelivery action.
        """
        arguments = {
            "ItemDeliveryMethod": ItemDeliveryMethod,
            "Expires": Expires,
        }

        self.call_action("RegisterItemDelivery", arguments=arguments, aspects=aspects)

        return

    def action_RegisterTelCPName(self, TelCPName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterTelCPName action.
        """
        arguments = {
            "TelCPName": TelCPName,
        }

        self.call_action("RegisterTelCPName", arguments=arguments, aspects=aspects)

        return

    def action_UnregisterTelCPName(self, TelCPName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UnregisterTelCPName action.
        """
        arguments = {
            "TelCPName": TelCPName,
        }

        self.call_action("UnregisterTelCPName", arguments=arguments, aspects=aspects)

        return

    def action_UpdateCalendarItems(self, Caltems, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdateCalendarItems action.
        """
        arguments = {
            "Caltems": Caltems,
        }

        self.call_action("UpdateCalendarItems", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ZoneGroupTopology1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'ZoneGroupTopology1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:ZoneGroupTopology:1'

    SERVICE_DEFAULT_VARIABLES = {
        "DiagnosticID": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "AlarmRunSequence": { "data_type": "string", "default": None, "allowed_list": None},
        "AreasUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "AvailableSoftwareUpdate": { "data_type": "string", "default": None, "allowed_list": None},
        "MuseHouseholdId": { "data_type": "string", "default": None, "allowed_list": None},
        "NetsettingsUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "SourceAreasUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "ThirdPartyMediaServersX": { "data_type": "string", "default": None, "allowed_list": None},
        "ZoneGroupID": { "data_type": "string", "default": None, "allowed_list": None},
        "ZoneGroupName": { "data_type": "string", "default": None, "allowed_list": None},
        "ZoneGroupState": { "data_type": "string", "default": None, "allowed_list": None},
        "ZonePlayerUUIDsInGroup": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_BeginSoftwareUpdate(self, UpdateURL, Flags, ExtraOptions, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the BeginSoftwareUpdate action.
        """
        arguments = {
            "UpdateURL": UpdateURL,
            "Flags": Flags,
            "ExtraOptions": ExtraOptions,
        }

        self.call_action("BeginSoftwareUpdate", arguments=arguments, aspects=aspects)

        return

    def action_CheckForUpdate(self, UpdateType, CachedOnly, Version, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CheckForUpdate action.

            :returns: "UpdateItem"
        """
        arguments = {
            "UpdateType": UpdateType,
            "CachedOnly": CachedOnly,
            "Version": Version,
        }

        out_params = self.call_action("CheckForUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UpdateItem",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetZoneGroupAttributes(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetZoneGroupAttributes action.

            :returns: "CurrentZoneGroupName", "CurrentZoneGroupID", "CurrentZonePlayerUUIDsInGroup", "CurrentMuseHouseholdId"
        """
        arguments = { }

        out_params = self.call_action("GetZoneGroupAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentZoneGroupName", "CurrentZoneGroupID", "CurrentZonePlayerUUIDsInGroup", "CurrentMuseHouseholdId",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetZoneGroupState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetZoneGroupState action.

            :returns: "ZoneGroupState"
        """
        arguments = { }

        out_params = self.call_action("GetZoneGroupState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ZoneGroupState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RegisterMobileDevice(self, MobileDeviceName, MobileDeviceUDN, MobileIPAndPort, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RegisterMobileDevice action.
        """
        arguments = {
            "MobileDeviceName": MobileDeviceName,
            "MobileDeviceUDN": MobileDeviceUDN,
            "MobileIPAndPort": MobileIPAndPort,
        }

        self.call_action("RegisterMobileDevice", arguments=arguments, aspects=aspects)

        return

    def action_ReportAlarmStartedRunning(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReportAlarmStartedRunning action.
        """
        arguments = { }

        self.call_action("ReportAlarmStartedRunning", arguments=arguments, aspects=aspects)

        return

    def action_ReportUnresponsiveDevice(self, DeviceUUID, DesiredAction, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ReportUnresponsiveDevice action.
        """
        arguments = {
            "DeviceUUID": DeviceUUID,
            "DesiredAction": DesiredAction,
        }

        self.call_action("ReportUnresponsiveDevice", arguments=arguments, aspects=aspects)

        return

    def action_SubmitDiagnostics(self, IncludeControllers, Type, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SubmitDiagnostics action.

            :returns: "DiagnosticID"
        """
        arguments = {
            "IncludeControllers": IncludeControllers,
            "Type": Type,
        }

        out_params = self.call_action("SubmitDiagnostics", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DiagnosticID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: waiting
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains framework timeout constants.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, Callable, Dict, List, Optional, Protocol

import os
import threading
import time

from datetime import datetime, timedelta

from rich.progress import Progress, BarColumn, TextColumn, TimeRemainingColumn
from rich.table import Column

from akit.comparisons import compare_any_of_same_type

from akit.timeouts import (
    TimeoutContext,
    TimeoutState,
    DEFAULT_WAIT_DELAY,
    DEFAULT_WAIT_INTERVAL,
    DEFAULT_WAIT_TIMEOUT,
    DEFAULT_WHATFOR_TEMPLATE
)

class WaitContext(TimeoutContext):
    """
        Place holder for differences that might arise between the base TimeoutContext and
        the WaitContext used for wait loops.
    """

class WaitCallback(Protocol):
    def __call__(self, wctx: TimeoutContext, *args, **kwargs) -> bool:
        """
            This specifies a callable object that can have variable arguments but
            that must have a final_attempt keywork arguement.  The expected behavior
            of the callback is to return false if the expected condition has not
            been meet.
        """


def wait_for_it(looper: WaitCallback, *largs, what_for: Optional[str]=None, delay: float=DEFAULT_WAIT_DELAY,
                interval: float=DEFAULT_WAIT_INTERVAL, timeout: float=DEFAULT_WAIT_TIMEOUT,
                lkwargs: Dict[Any, Any]={}, wctx: Optional[WaitContext]=None):
    """
        Provides for convenient mechanism to wait for criteria to be met before proceeding.

        :param looper: A callback method that is repeatedly called while it returns `False` up-to
            the end of a timeout period, and that will return `True` if a waited on condition is
            met prior to a timeout condition being met.
        :param largs: Arguements to pass to the looper callback function.
        :param what_for: A breif description of what is being waited for.
        :param delay: An initial time delay to consume before beginning the waiting process.
        :param interval: A period of time to delay between rechecks of the wait conditon
        :param timeout: The maximum period of time in seconds that should be waited before timing out.
        :param lkwargs: Additional keyword arguments to pass to the looper function

        :raises AKitTimeoutError: A timeout error with details around the wait condition.

        ..note: The 'delay', 'interval' and 'timeout' parameters will be ignored if the 'wctx' parameter
                is passed as the wctx (WaitContext) parameter includes these values with it.
    """

    if what_for is None:
        what_for = DEFAULT_WHATFOR_TEMPLATE.format(looper.__name__)

    if wctx is None:
        wctx = WaitContext(timeout, interval=interval, delay=delay, what_for=what_for)

    if wctx.delay > 0:
        time.sleep(DEFAULT_WAIT_DELAY)

    wctx.mark_begin()

    condition_met = False

    while True:
        condition_met = looper(wctx, *largs, **lkwargs)
        if condition_met:
            wctx.mark_complete()
            break

        if not wctx.should_continue():
            break

        if wctx.interval > 0:
            time.sleep(wctx.interval)

    if not condition_met:
        # Mark the time we are performing the final attempt
        wctx.mark_time()
        wctx.mark_final_attempt()
        condition_met = looper(wctx, *largs, **lkwargs)

    if not condition_met:
        toerr = wctx.create_timeout(what_for)
        raise toerr

    wctx.mark_complete()

    return


class WaitGate:

    def __init__(self, gate: threading.Event, message: Optional[str]=None, timeout: Optional[float]=None,
                 timeout_args: Optional[list]=None):
        self._gate = gate
        self._messaage = message
        self._timeout = timeout
        self._timeout_args = timeout_args
        return

    @property
    def gate(self) -> threading.Event:
        return self._gate

    @property
    def message(self) -> str:
        return self._message

    @property
    def timeout(self) -> float:
        return self._timeout

    @property
    def timeout_args(self) -> list:
        return self._timeout_args

    def clear(self):
        self._gate.clear()
        return

    def is_set(self) -> bool:
        rtnval = self._gate.is_set()
        return rtnval

    def set(self):
        self._gate.set()
        return

    def wait(self, timeout: Optional[float]=None, raise_timeout=False):

        if timeout is None:
            timeout = self._timeout

        rtnval = self._gate.wait(timeout=self._timeout)
        if not rtnval:
            errmsg = ""
            raise TimeoutError(errmsg)

        return rtnval

class WaitingScope:
    def __init__(self, gates: List[WaitGate],):
        self._gates = gates
        return

    def __enter__(self):
        for gate in self._gates:
            gate.clear()
        return
    
    def __exit__(self, ex_type, ex_inst, ex_tb):
        return
    
    def wait(self):

        for gate in self.gates:
            gate.wait()

        return

class MultiEvent:

    def __init__(self, contexts: List[object]):
        self._contexts = contexts
        return


class ProgressOfDurationWaitContext(WaitContext):


    def __init__(self, progress: Progress, task_name: str, timeout: float, delay: float=0, what_for: Optional[str]=None):
        WaitContext.__init__(self, timeout, delay=delay, what_for=what_for)
        self._progress = progress
        self._time_delta_secs = None

        self._task_name = task_name
        self._task_id = None
        return

    def mark_begin(self):
        """
            Mark the wait context as running.
        """

        self._now_time = datetime.now()
        self._start_time = self._now_time
        self._end_time = self._start_time + timedelta(seconds=self._timeout)

        self._time_delta_secs = (self._end_time - self._start_time).total_seconds()
        self._task_id = self._progress.add_task(self._task_name, total=self._time_delta_secs)
        self._progress.update(self._task_id, completed=0)

        self._wait_state = TimeoutState.Running
        return

    def mark_complete(self):
        """
            Mark the wait context as complete.
        """
        self._wait_state = TimeoutState.Completed
        self.render_complete()
        return

    def mark_timeout(self):
        """
            Called to mark the wait context as timed out.
        """
        self._wait_state = TimeoutState.TimedOut
        self.render_timeout()
        return

    def render_complete(self):
        if self._task_id is not None:
            self._progress.update(self._task_id, completed=self._time_delta_secs)
        return

    def render_progress(self):
        current_secs = (self._now_time - self._start_time).total_seconds()

        if self._task_id is not None:
            self._progress.update(self._task_id, completed=current_secs)

        return

    def render_timeout(self):
        self._progress.update(self._task_id, completed=self._time_delta_secs)
        return

    def should_continue(self) -> bool:
        """
            Indicates if a wait condition should continue based on time specifications and context.
        """
        self._now_time = datetime.now()

        scont = True

        if self._wait_state == TimeoutState.Completed:
            scont = False
        elif self._now_time > self._end_time:
            scont = False

        self.render_progress()

        return scont

    def create_rich_progress() -> Progress:
        desc_column = TextColumn("{task.description}", table_column=Column(ratio=1))
        bar_column = BarColumn(bar_width=None, table_column=Column(ratio=2))
        time_remaining = TimeRemainingColumn()
        progress = Progress(desc_column, bar_column, "[progress.percentage]{task.percentage:>3.0f}%", time_remaining, expand=True)
        return progress

def wait_for_result_from_call(wctx: WaitContext, callable: Callable, expected: Any, *args, **kwargs):
    """
        Wait for the 'callable' provided to return the expeted result.
    """

    found = callable(*args, **kwargs)

    comp_result = compare_any_of_same_type(expected, found)
    if not comp_result and wctx.final_attempt:
        callable_name = callable.__name__
        whatfor = "for {} to return the expected result {}.  last={}".format(callable_name, expected, found)
        toerr = wctx.create_timeout(whatfor)
        raise toerr

    return comp_result



"""
.. module:: upnpprotocols
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`MSearchRootDeviceProtocol` class and
               associated diagnostic.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Tuple

import copy
import os
import re
import socket
import threading

import netifaces

from akit.exceptions import AKitTimeoutError

from akit.xlogging.foundations import getAutomatonKitLogger

from akit.networking.multicast import create_multicast_socket_for_iface, create_multicast_socket
from akit.networking.unicast import create_unicast_socket
from akit.networking.interfaces import get_interface_for_ip

from akit.waiting import ProgressOfDurationWaitContext, WaitContext

REGEX_NOTIFY_HEADER = re.compile("NOTIFY[ ]+[*/]+[ ]+HTTP/1")

logger = getAutomatonKitLogger()

class MSearchTargets:
    """
        MSearch target constants.
    """
    ROOTDEVICE = "upnp:rootdevice"
    ALL="ssdp:all"

class MSearchKeys:
    """
        MSearch dictionary or header keys that MSearch response data items are stored under.
    """
    CACHE_CONTROL = "CACHE-CONTROL"
    EXT = "EXT"
    LOCATION = "LOCATION"
    NTS = "NTS"
    SERVER = "SERVER"
    ST = "ST"
    USN = "USN"
    USN_DEV = "USN_DEV"
    USN_CLS = "USN_CLS"

    IP = "IP"
    ROUTES = "ROUTES"

class MSearchRouteKeys:
    """
        The dictionary keys that MSearch interface routing information is stored under.
    """
    IFNAME = "LOCAL_IFNAME"
    IP = "LOCAL_IP"

class UpnpProtocol:
    """
        UPnP Protocol constants.
    """
    MULTICAST_ADDRESS = '239.255.255.250'
    PORT = 1900

    HEADERS = {
        "ST": MSearchTargets.ROOTDEVICE,
        "Man": "ssdp:discover",
        "MX": "1"
    }

SSDP_IPV6_MULTICAST_ADDRESS_LINK_LOCAL="FF02::C"
SSDP_IPV6_MULTICAST_ADDRESS_SITE_LOCAL="FF05::C"
SSDP_IPV6_MULTICAST_ADDRESS_ORG_LOCAL="FF08::C"
SSDP_IPV6_MULTICAST_ADDRESS_GLOBAL_LOCAL="FF0E::"

class MSearchScanContext:
    """
        The :class:`MSearchScanContext` is an object that allows the sharing of scan results across threads
        that are scanning mulitple interfaces.  We utilize a shared scan context so we can find the devices
        faster.
    """

    def __init__(self, upnp_device_hints):
        self.upnp_device_hints = upnp_device_hints
        self.remaining_device_hints = [dh for dh in upnp_device_hints]

        self.have_hints = False
        if len(self.remaining_device_hints) > 0:
            self.have_hints = True

        self.continue_scan = True

        self.found_devices = {}
        self.matching_devices = {}

        self.lock = threading.Lock()
        return

    def register_device(self, ifname: str, usn: str, device_info: dict, route_info: dict):
        """
            Method called by msearch scan threads to register msearch results for a specific network interface.

            :param ifname: The interface name to register msearch results for.
            :param usn: The UPNP usn to register msearch results for.
            :param device_info: A dictionary containing the information about the device found by the msearch thread.
            :param route_info: The route info about the interface that the device was found on.
        """
        # pylint: disable=unused-argument
        self.lock.acquire()
        try:
            if usn not in self.found_devices:

                device_info[MSearchKeys.ROUTES] = {ifname: route_info}

                self.found_devices[usn] = device_info

                for hint in self.remaining_device_hints:
                    if usn.find(hint) > -1:
                        self.remaining_device_hints.remove(hint)
                        self.matching_devices[hint] = device_info
                        break

                if self.have_hints and len(self.remaining_device_hints) == 0:
                    self.continue_scan = False
            else:
                existing_info = self.found_devices[usn]

                if MSearchKeys.ROUTES not in existing_info:
                    existing_info[MSearchKeys.ROUTES] = {ifname: route_info}
                else:
                    existing_info[MSearchKeys.ROUTES].update({ifname: route_info})

        finally:
            self.lock.release()

        return

def msearch_parse_request(content: bytes) -> dict:
    """
        Takes in the content of the MSEARCH request and parses it into a
        python dictionary object.

        :param content: MSearch request content as a bytes.

        :return: A python dictionary with key and values from the MSearch request
    """
    content = content.decode('utf-8')

    respinfo = None

    resplines = content.splitlines(False)
    if len(resplines) > 0:
        header = resplines.pop(0).strip()
        if header.startswith("M-SEARCH *"):
            respinfo = {}
            for nxtline in resplines:
                cidx = nxtline.find(":")
                if cidx > -1:
                    key = nxtline[:cidx].upper()
                    val = nxtline[cidx+1:].strip()
                    respinfo[key] = val

    return respinfo

def msearch_parse_response(content: bytes) -> dict:
    """
        Takes in the content of the response of an MSEARCH request and parses it into a
        python dictionary object.

        :param content: MSearch response content as a bytes.

        :return: A python dictionary with key and values from the MSearch response
    """
    content = content.decode('utf-8')

    respinfo = None

    resplines = content.splitlines(False)
    if len(resplines) > 0:
        header = resplines.pop(0).strip()
        if header.startswith("HTTP/"):
            respinfo = {}
            for nxtline in resplines:
                cidx = nxtline.find(":")
                if cidx > -1:
                    key = nxtline[:cidx].upper()
                    val = nxtline[cidx+1:].strip()
                    respinfo[key] = val

    return respinfo


def msearch_query_host(target_address: str, query_usn: Optional[str]=None, mx: int = 5, st: str = MSearchTargets.ROOTDEVICE,
                response_timeout: float = 45, ttl: int = 1, unicast_socket: bool=False, custom_headers: Optional[dict]=None,
                wctx: Optional[WaitContext]=None):
    """
        The inline msearch function provides a mechanism to do a synchronous msearch
        in order to determine if a specific host  devices is available and to
        determine the interface that the device is available on.

        :param target_address: The network address of the device being targeted by the unicast mquery
        :param query_usn: The USN of the device that is being queried
        :param mx:  Instructs the M-SEARCH targets to wait a random time from 0-mx before sending a response.
        :param st: The search target of the MSearch.
        :param response_timeout:  The timeout to wait for responses from all the expected devices.
        :param ttl: The time to live for the multicast packet
                    0 = same host
                    1 = same subnet
                    32 = same site
                    64 = same region
                    128 = same continent
                    255 = unrestricted scope
        :param unicast_socket: Utilize a unicast socket configuration for the query, else use a multicast socket configuration
        :param custom_headers: Custom headers that are added to the query message.

        :raises: TimeoutError, KeyboardInterrupt
    """

    found_device_info = None

    target_address = target_address.encode()

    msearch_msg_lines = [
        b'M-SEARCH * HTTP/1.1',
        b'HOST: %s:%d' % (UpnpProtocol.MULTICAST_ADDRESS.encode("utf-8"), UpnpProtocol.PORT),
        b'MAN: "ssdp:discover"',
        b'ST: %s' % st.encode("utf-8")
    ]

    if custom_headers is not None:
        for hname, hval in custom_headers.items():
            hname = hname.upper().encode("utf-8")
            msearch_msg_lines.append(b'%s: %s' % (hname, hval.encode("utf-8")))

    msearch_msg_lines.append(b'')

    msearch_msg = b"\r\n".join(msearch_msg_lines)

    sock = None

    if unicast_socket:
        sock = create_unicast_socket(target_address, UpnpProtocol.PORT, socket.AF_INET, ttl=ttl, timeout=response_timeout)
    else:
        sock = create_multicast_socket(UpnpProtocol.MULTICAST_ADDRESS, UpnpProtocol.PORT, socket.AF_INET, ttl=ttl, timeout=response_timeout)
                
    if wctx is None:
        wctx = WaitContext(response_timeout)

    wctx.mark_begin()

    try:
        sock.sendto(msearch_msg, (target_address, UpnpProtocol.PORT))

        while True:
     
            try:
                resp, addr = sock.recvfrom(1024)

                if addr == target_address:
                    if resp.startswith(b"NOTIFY * HTTP/"):
                        resp = resp.lstrip(b"NOTIFY * ")

                    if resp.startswith(b"HTTP/"):
                        device_info = msearch_parse_response(resp)
                        if device_info is not None:
                            foundst = device_info.get(MSearchKeys.NTS, None)
                            if foundst == 'ssdp:alive':
                                if MSearchKeys.USN in device_info:
                                    usn_dev= device_info[MSearchKeys.USN]
                                    usn_dev = usn_dev.lstrip("uuid:")
                                    
                                    device_info[MSearchKeys.USN_DEV] = usn_dev

                                    if query_usn is None or usn_dev == query_usn:

                                        ip_addr = addr[0]
                                        ifname = get_interface_for_ip(ip_addr)

                                        route_info = {
                                            MSearchRouteKeys.IFNAME: ifname,
                                            MSearchRouteKeys.IP: ip_addr
                                        }
                                        device_info[MSearchKeys.ROUTES] = {ifname: route_info}
                                        device_info[MSearchKeys.IP] = ip_addr

                                        found_device_info = device_info

                                else:
                                    print("device_info didn't have a USN. %r" % device_info)
                        else:
                            print("device_info was None.")
                else:
                    print("Respons from other addr={}".format(addr))

            except socket.timeout as toerr:
                pass
            
            if not wctx.should_continue() and wctx.has_timed_out:
                wctx.mark_timeout()
                break
    except Exception:
        logger.exception("msearch_query_host exception:")
    except KeyboardInterrupt: # pylint: disable=try-except-raise
        raise
    finally:
        sock.close()

    return found_device_info


def msearch_on_interface(scan_context: MSearchScanContext, ifname: str, ifaddress: str, *,
                         mx: int = 5, st: str = MSearchTargets.ROOTDEVICE, response_timeout: float = 45,
                         ttl: int = 1, custom_headers: Optional[dict]=None, wctx: Optional[WaitContext]=None):
    """
        The inline msearch function provides a mechanism to do a synchronous msearch
        in order to determine if a set of available devices are available and to
        determine the interfaces that each device is listening on.

        :param scan_context: A shared scan context that shares information between the scan threads and
                             speeds up the process of finding the expected UPNP devices across the
                             interfaces.
        :param ifname: The name if the interface being searched.
        :param ifaddress: The IP address of the interface being searched.
        :param mx:  Instructs the M-SEARCH targets to wait a random time from 0-mx before sending a response.
        :param st: The search target of the MSearch.
        :param response_timeout:  The timeout to wait for responses from all the expected devices.
        :param interval: The retry interval to wait before retrying to search for an expected device.
        :param ttl: The time to live for the multicast packet
                    0 = same host
                    1 = same subnet
                    32 = same site
                    64 = same region
                    128 = same continent
                    255 = unrestricted scope
        :param custom_headers: Optional custom msearch headers.
        :param wctx: An option wait context to use for waiting

        :returns:  dict -- A dictionary of the devices that were found.
        :raises: TimeoutError, KeyboardInterrupt
    """

    route_info = {
        MSearchRouteKeys.IFNAME: ifname,
        MSearchRouteKeys.IP: ifaddress
    }

    multicast_address = UpnpProtocol.MULTICAST_ADDRESS
    multicast_port = UpnpProtocol.PORT

    msearch_msg_lines = [
        b'M-SEARCH * HTTP/1.1',
        b'HOST: %s:%d' % (UpnpProtocol.MULTICAST_ADDRESS.encode("utf-8"), UpnpProtocol.PORT),
        b'MAN: "ssdp:discover"',
        b'ST: %s' % st.encode("utf-8")
    ]

    if custom_headers is not None:
        for hname, hval in custom_headers.items():
            hname = hname.upper().encode("utf-8")
            msearch_msg_lines.append(b'%s: %s' % (hname, hval.encode("utf-8")))

    msearch_msg_lines.append(b'')

    msearch_msg = b"\r\n".join(msearch_msg_lines)

    sock = None

    try:
        sock = create_multicast_socket_for_iface(UpnpProtocol.MULTICAST_ADDRESS, ifname, UpnpProtocol.PORT, socket.AF_INET, ttl=ttl, timeout=5)

        sock.sendto(msearch_msg, (multicast_address, multicast_port))

        if wctx is None:
            wctx = WaitContext(response_timeout)

        wctx.mark_begin()

        while True:

            try:
                resp, addr = sock.recvfrom(1024)
                if resp.startswith(b"NOTIFY * HTTP/"):
                    resp = resp.lstrip(b"NOTIFY * ")

                if resp.startswith(b"HTTP/"):
                    device_info = msearch_parse_response(resp)

                    if device_info is not None:
                        foundst = device_info.get(MSearchKeys.ST, None)
                        if foundst == st:
                            if MSearchKeys.USN in device_info:
                                usn_dev, usn_cls = device_info[MSearchKeys.USN].split("::")
                                usn_dev = usn_dev.lstrip("uuid:")
                                if usn_cls == "upnp:rootdevice":
                                    device_info[MSearchKeys.USN_DEV] = usn_dev
                                    device_info[MSearchKeys.USN_CLS] = usn_cls

                                    device_info[MSearchKeys.IP] = addr[0]

                                    scan_context.register_device(ifname, usn_dev, device_info, route_info)
                            else:
                                print("device_info didn't have a USN. %r" % device_info)
                    else:
                        print("device_info was None.")

            except socket.timeout:
                pass

            if not wctx.should_continue() and wctx.has_timed_out:
                wctx.mark_timeout()
                break

            if not scan_context.continue_scan:
                break
    except Exception:
        logger.exception("msearch_on_interface exception:")
    except KeyboardInterrupt: # pylint: disable=try-except-raise
        raise
    finally:
        sock.close()

    return


def msearch_scan(expected_devices, interface_list=None, response_timeout=45, interval=2, raise_exception=False,
                 custom_headers: Optional[dict]=None, show_progress=False) -> Tuple[dict, dict]:
    """
        Performs a msearch across a list of interfaces for a specific expected device.  This method is typically used
        during a persistent search when a device was not found in a broad msearch.

        :param expected_devices: A list of USN(s) of a list of devices to search for.
        :param interface_list: A list of interface names to scan for the device.
        :param response_timeout:  The timeout to wait for responses from all the expected devices.
        :param interval: The retry interval to wait before retrying to search for an expected device.
        :param raise_exception: A boolean indicating if the mquery should raise an exception on failure.
        :param custom_headers: Custom headers to include in the msearch message
        :param show_progress: Should progress be rendered to the console.

        :returns: A tuple with a dictionary of found and a dictionary of matching devices found.
    """
    if interface_list is None:
        interface_list = netifaces.interfaces()

    scan_context = MSearchScanContext(expected_devices)

    search_threads = []

    progress = None
    if show_progress:
        progress = ProgressOfDurationWaitContext.create_rich_progress()

    try:
        if progress is not None:
            progress.start()

        for ifname in interface_list:
            ifaddress = None

            address_info = netifaces.ifaddresses(ifname)
            if address_info is not None:
                # First look for IPv4 address information
                if netifaces.AF_INET in address_info:
                    addr_info = address_info[netifaces.AF_INET][0]
                    ifaddress = addr_info["addr"]

                # If we didn't find an ipv4 address, try using IPv6
                # if ifaddress is None and netifaces.AF_INET6 in addr_info:
                #    addr_info = address_info[netifaces.AF_INET6][0]
                #    ifaddress = addr_info["addr"]

                if ifaddress is not None:
                    thname = "msearch-%s" % ifname
                    thargs = (scan_context, ifname, ifaddress)
                    wctx = None
                    if progress is not None:
                        wctx = ProgressOfDurationWaitContext(progress, thname, timeout=response_timeout)
                    else:
                        wctx = WaitContext(timeout=response_timeout)

                    thkwargs = { "wctx": wctx, "custom_headers": custom_headers}

                    sthread = threading.Thread(name=thname, target=msearch_on_interface, args=thargs, kwargs=thkwargs)
                    sthread.start()
                    search_threads.append(sthread)

        # Wait for all the search threads to finish
        while len(search_threads) > 0:
            nxt_thread = search_threads.pop(0)
            nxt_thread.join()

    finally:
        if progress is not None:
            progress.stop()

    found_devices = copy.deepcopy(scan_context.found_devices)
    matching_devices = copy.deepcopy(scan_context.matching_devices)

    if raise_exception and len(matching_devices) != len(expected_devices):
        err_msg = "Failed to find expected UPNP devices after a timeout of %s seconds.\n" % response_timeout

        missing = [dkey for dkey in expected_devices]
        for dkey in scan_context.found_devices:
            if dkey in missing:
                missing.remove(dkey)

        err_msg_lines = []
        err_msg_lines.append("EXPECTED: (%s)" % len(expected_devices))
        for dkey in expected_devices:
            err_msg_lines.append("    %r:" % dkey)
        err_msg_lines.append("")

        err_msg_lines.append("MATCHING: (%s)" % len(matching_devices))
        for dkey in matching_devices:
            err_msg_lines.append("    %r:" % dkey)
        err_msg_lines.append("")

        err_msg_lines.append("FOUND: (%s)" % len(found_devices))
        for dkey in found_devices:
            err_msg_lines.append("    %r:" % dkey)
        err_msg_lines.append("")

        err_msg_lines.append("MISSING: (%s)" % len(missing))
        for dkey in missing:
            err_msg_lines.append("    %r:" % dkey)
        err_msg_lines.append("")

        err_msg = os.linesep.join(err_msg_lines)
        raise AKitTimeoutError(err_msg) from None

    return found_devices, matching_devices

def notify_parse_request(content: str) -> dict:
    """
        Takes in the content of the NOTIFY request and parses it into a
        python dictionary object.

        :param content: Notify request content as a string.

        :return: A python dictionary with key and values from the Notify request
    """
    content = content.decode('utf-8')

    resp_headers = None
    resp_body = None

    mobj = REGEX_NOTIFY_HEADER.search(content)
    if mobj is not None:
        header_content = None
        body_content = None

        if content.find("\r\n\r\n") > -1:
            header_content, body_content = content.split("\r\n\r\n", 1)
        elif content.find("\n\n") > -1:
            header_content, body_content = content.split("\n\n", 1)
        else:
            header_content = content

        resplines = header_content.splitlines(False)

        # Pop the NOTIFY header
        resplines.pop(0).strip()

        resp_headers = {}
        for nxtline in resplines:
            cidx = nxtline.find(":")
            if cidx > -1:
                key = nxtline[:cidx].upper()
                val = nxtline[cidx+1:].strip()
                resp_headers[key] = val

        if body_content is not None:
            resp_body = body_content

    return resp_headers, resp_body
